"""
Classes for handling raster data.
"""

from . import grid
from . import aaigrid
from . import raster

from .grid import RegularGrid, WarpedGrid
from .read import read_aai, read_gtiff, aairead, gtiffread
from .aaigrid import AAIGrid
from .raster import witch_of_agnesi, peaks, pad, slope, aspect, grad, div
from .raster import normed_vector_field

try:
    from .crfuncs import streamline2d
except ImportError:
    from .streamline import streamline2d

__all__ = ["grid", "aaigrid", "raster",
           "RegularGrid", "WarpedGrid",
           "aairead", "gtiffread", "read_aai", "read_gtiff",
           "AAIGrid",
           "pad", "slope", "aspect", "grad", "div", "normed_vector_field",
           "streamline2d"]

