from .vector import Point, Multipoint, Line, Polygon
from .crs import LonLatWGS84

us_capitols = [Point((-112.1, 33.57), properties={"n": "Phoenix, Arizona, United States"}, crs=LonLatWGS84),
               Point((-121.5, 38.57), properties={"n": "Sacramento, California, United States"}, crs=LonLatWGS84),
               Point((-84.42, 33.76), properties={"n": "Atlanta, Georgia, United States"}, crs=LonLatWGS84),
               Point((-86.15, 39.78), properties={"n": "Indianapolis, Indiana, United States"}, crs=LonLatWGS84),
               Point((-112.0, 46.60) , properties={"n": "Helena, Montana, United States"}, crs=LonLatWGS84),
               Point((-82.99, 39.98), properties={"n": "Columbus, Ohio, United States"}, crs=LonLatWGS84),
               Point((-77.48, 37.53), properties={"n": "Richmond, Virginia, United States"}, crs=LonLatWGS84),
               Point((-95.69, 39.04), properties={"n": "Topeka, Kansas, United States"}, crs=LonLatWGS84),
               Point((-71.02, 42.33), properties={"n": "Boston, Massachusetts, United States"}, crs=LonLatWGS84),
               Point((-96.68, 40.81), properties={"n": "Lincoln, Nebraska, United States"}, crs=LonLatWGS84),
               Point((-97.51, 35.47), properties={"n": "Oklahoma City, Oklahoma, United States"}, crs=LonLatWGS84),
               Point((-134.2, 58.37), properties={"n": "Juneau, Alaska, United States"}, crs=LonLatWGS84),
               Point((-100.3, 44.38), properties={"n": "Pierre, South Dakota, United States"}, crs=LonLatWGS84),
               Point((-157.8, 21.31), properties={"n": "Honolulu, Hawaii, United States"}, crs=LonLatWGS84),
               Point((-86.27, 32.35), properties={"n": "Montgomery, Alabama, United States"}, crs=LonLatWGS84),
               Point((-92.36, 34.73), properties={"n": "Little Rock, Arkansas, United States"}, crs=LonLatWGS84),
               Point((-104.9, 39.76), properties={"n": "Denver, Colorado, United States"}, crs=LonLatWGS84),
               Point((-72.68, 41.77), properties={"n": "Hartford, Connecticut, United States"}, crs=LonLatWGS84),
               Point((-75.52, 39.16), properties={"n": "Dover, Delaware, United States"}, crs=LonLatWGS84),
               Point((-77.02, 38.90) , properties={"n": "Washington, District of Columbia, United States"}, crs=LonLatWGS84),
               Point((-84.25, 30.46), properties={"n": "Tallahassee, Florida, United States"}, crs=LonLatWGS84),
               Point((-116.2, 43.60) , properties={"n": "Boise, Idaho, United States"}, crs=LonLatWGS84),
               Point((-89.67, 39.76), properties={"n": "Springfield, Illinois, United States"}, crs=LonLatWGS84),
               Point((-93.62, 41.57), properties={"n": "Des Moines, Iowa, United States"}, crs=LonLatWGS84),
               Point((-84.87, 38.19), properties={"n": "Frankfort, Kentucky, United States"}, crs=LonLatWGS84),
               Point((-91.13, 30.45), properties={"n": "Baton Rouge, Louisiana, United States"}, crs=LonLatWGS84),
               Point((-69.73, 44.33), properties={"n": "Augusta, Maine, United States"}, crs=LonLatWGS84),
               Point((-76.51, 38.97), properties={"n": "Annapolis, Maryland, United States"}, crs=LonLatWGS84),
               Point((-84.56, 42.71), properties={"n": "Lansing, Michigan, United States"}, crs=LonLatWGS84),
               Point((-93.10, 44.95), properties={"n": "Saint Paul, Minnesota, United States"}, crs=LonLatWGS84),
               Point((-90.21, 32.32), properties={"n": "Jackson, Mississippi, United States"}, crs=LonLatWGS84),
               Point((-92.18, 38.57), properties={"n": "Jefferson City, Missouri, United States"}, crs=LonLatWGS84),
               Point((-119.7, 39.15), properties={"n": "Carson City, Nevada, United States"}, crs=LonLatWGS84),
               Point((-71.56, 43.23), properties={"n": "Concord, New Hampshire, United States"}, crs=LonLatWGS84),
               Point((-74.76, 40.22), properties={"n": "Trenton, New Jersey, United States"}, crs=LonLatWGS84),
               Point((-106.0, 35.67), properties={"n": "Santa Fe, New Mexico, United States"}, crs=LonLatWGS84),
               Point((-73.80, 42.67), properties={"n": "Albany, New York, United States"}, crs=LonLatWGS84),
               Point((-78.64, 35.83), properties={"n": "Raleigh, North Carolina, United States"}, crs=LonLatWGS84),
               Point((-100.8, 46.81), properties={"n": "Bismarck, North Dakota, United States"}, crs=LonLatWGS84),
               Point((-123.0, 44.92), properties={"n": "Salem, Oregon, United States"}, crs=LonLatWGS84),
               Point((-76.89, 40.28), properties={"n": "Harrisburg, Pennsylvania, United States"}, crs=LonLatWGS84),
               Point((-71.42, 41.82), properties={"n": "Providence, Rhode Island, United States"}, crs=LonLatWGS84),
               Point((-81.01, 34.02), properties={"n": "Columbia, South Carolina, United States"}, crs=LonLatWGS84),
               Point((-86.78, 36.17), properties={"n": "Nashville, Tennessee, United States"}, crs=LonLatWGS84),
               Point((-97.76, 30.31), properties={"n": "Austin, Texas, United States"}, crs=LonLatWGS84),
               Point((-111.9, 40.78), properties={"n": "Salt Lake City, Utah, United States"}, crs=LonLatWGS84),
               Point((-72.57, 44.27), properties={"n": "Montpelier, Vermont, United States"}, crs=LonLatWGS84),
               Point((-122.9, 47.04), properties={"n": "Olympia, Washington, United States"}, crs=LonLatWGS84),
               Point((-81.63, 38.35), properties={"n": "Charleston, West Virginia, United States"}, crs=LonLatWGS84),
               Point((-89.43, 43.09), properties={"n": "Madison, Wisconsin, United States"}, crs=LonLatWGS84),
               Point((-104.8, 41.15), properties={"n": "Cheyenne, Wyoming, United States"}, crs=LonLatWGS84)]
